import pytest
import numpy as np
from tsdistances import (
    euclidean_distance,
    erp_distance,
    dtw_distance,
    adtw_distance,
    twe_distance,
)
from aeon.distances import (
    erp_pairwise_distance,
    dtw_pairwise_distance,
    adtw_pairwise_distance,
    twe_pairwise_distance,
)
import time
import pandas as pd
import pathlib

UCR_ARCHIVE_PATH = pathlib.Path('../../DATA/ucr')
BENCHMARKS_DS = ["ACSF1", "Adiac", "Beef", "CBF", "ChlorineConcentration", "CinCECGTorso", "CricketX", "DiatomSizeReduction", "DistalPhalanxOutlineCorrect", "ECG200", "EthanolLevel", "FreezerRegularTrain", "FreezerSmallTrain", "Ham", "Haptics", "HouseTwenty", "ItalyPowerDemand", "MixedShapesSmallTrain", "NonInvasiveFetalECGThorax1", "ShapesAll", "Strawberry", "UWaveGestureLibraryX", "Wafer"]
# TSDISTANCES = [erp_distance, adtw_distance]
# AEONDISTANCES = [erp_pairwise_distance, adtw_pairwise_distance]
TSDISTANCES = [dtw_distance]
AEONDISTANCES = [dtw_distance]
MODALITIES = ["", "par", "gpu"]

def load_benchmark():
    benchmark_ds = sorted([x for x in UCR_ARCHIVE_PATH.iterdir() if x.name in BENCHMARKS_DS])
    return benchmark_ds

DATASETS_PATH = load_benchmark()

def test_draw_scatter_ucr():
    import matplotlib.pyplot as plt
    import seaborn as sns
    import pandas as pd

    ucr_datasets = sorted([x for x in UCR_ARCHIVE_PATH.iterdir() if x.is_dir()])
    ucr_info = np.zeros((len(ucr_datasets), 3), dtype=int)
    is_benchmark = np.empty(len(ucr_datasets), dtype=str)

    for i, dataset in enumerate(ucr_datasets):
        train = np.loadtxt(dataset / f"{dataset.name}_TRAIN.tsv", delimiter="\t")
        test = np.loadtxt(dataset / f"{dataset.name}_TEST.tsv", delimiter="\t")
        X_train, _ = train[:, 1:], train[:, 0]
        X_test, _ = test[:, 1:], test[:, 0]

        X = np.vstack((X_train, X_test))
        ucr_info[i] = [X_train.shape[0], X_test.shape[0], X.shape[1]]  # Total number of time series, Time series length, Time series length - 1
        is_benchmark[i] = "Benchmarked" if dataset.name in BENCHMARKS_DS else "Non-benchmarked"
    df = pd.DataFrame(np.column_stack([np.where(is_benchmark=='B')[0]+1, ucr_info[np.where(is_benchmark=='B')[0]]]), columns=["ID", "Train Size", "Test Size", "Time Series Length"], index=[ds.name for ds in DATASETS_PATH])
    df.to_latex("ucr_dataset_info.tex", index=True, float_format="%.0f", escape=False, column_format="lcccc", label='tab:ucr_datasets_info', caption="UCR Dataset Information. The table shows the number of time series in the training and test sets, as well as the length of the time series for each dataset.")
    # Create the scatter plot
    ds_size = ucr_info[:, :2].sum(axis=1)
    data = pd.DataFrame({"Dataset size": ds_size, "Time series Length": ucr_info[:, 2], "Benchmark Status": is_benchmark})

    sns.scatterplot(data=data[data["Benchmark Status"]=="N"], x='Dataset size', y='Time series Length', label='Non-Benchmarked', marker='o')
    sns.scatterplot(data=data[data["Benchmark Status"]=="B"], x='Dataset size', y='Time series Length', label='Benchmarked', marker='x', linewidth=2)

    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel("Dataset size (log scale)")
    plt.ylabel("Time series Length (log scale)")
    plt.legend()
    plt.title("UCR Archive Datasets")
    flag = True
    for i in range(len(ucr_datasets)):
        if is_benchmark[i] == "B":
            if i not in [43, 44]:  # Exclude the last two datasets for clarity
                plt.text(
                    ds_size[i],
                    ucr_info[i, 2],
                    str(i+1),
                    ha = 'center',
                    va = 'top',
                    color = 'black',
                    fontsize=6,
                )
            else:
                if flag:
                    plt.text(
                        ds_size[i],
                        ucr_info[i, 2],
                        f"[44-45]",
                        ha = 'center',
                        va = 'top',
                        color = 'black',
                        fontsize=6,
                    )
                    flag = False

    plt.savefig("benchmark_datasets.svg", dpi=300)

def test_with_threads():
    import matplotlib.pyplot as plt
    import seaborn as sns

    times = np.loadtxt("tests/ACSF1/times_per_thread.csv", delimiter=",")
    times = times.mean(axis=1)
    # bar plot with number of threads on x and times on y
    sns.barplot(x=np.arange(1, 17), y=times[::-1])
    plt.xlabel("Number of Threads")
    plt.ylabel("Time (s)")
    plt.title("ACSF1 vs Number of Threads")
    plt.savefig("tests/times_per_thread.svg", dpi=300)
    caption = "Elapsed time computing the DTW distance on the ACSF1 Dataset (TRAIN vs TEST), changing the number of threads used."

def test_tsdistances():
    tsdistances_times = np.full((len(DATASETS_PATH), len(TSDISTANCES), len(MODALITIES)), np.nan)
    aeon_times = np.full((len(DATASETS_PATH), len(TSDISTANCES)), np.nan)

    for i, dataset in enumerate(DATASETS_PATH):
        if dataset.name != "NonInvasiveFetalECGThorax1":
            continue
        print(f"\nDataset: {dataset.name}")
        train = np.loadtxt(dataset / f"{dataset.name}_TRAIN.tsv", delimiter="\t")
        test = np.loadtxt(dataset / f"{dataset.name}_TEST.tsv", delimiter="\t")
        X_train = train[:, 1:]
        X_test = test[:, 1:]

        for j, (tsdist, aeondist)  in enumerate(zip(TSDISTANCES, AEONDISTANCES)):
            start = time.time()
            D = tsdist(X_train, X_test, par=False)
            end = time.time()
            tsdistances_times[i, j, 0] = end - start

            start = time.time()
            D_par = tsdist(X_train, X_test, par=True)
            end = time.time()
            tsdistances_times[i, j, 1] = end - start

            if tsdist.__name__ != "euclidean_distance":
                start = time.time()
                D_gpu = tsdist(X_train, X_test, device='gpu')
                end = time.time()
                print(end - start)
                tsdistances_times[i, j, 2] = end - start
            # AEON distances
            start = time.time()
            D_aeon = aeondist(X_train, X_test)
            end = time.time()
            aeon_times[i, j] = end - start

            print(f"\t{tsdist.__name__} - \n\t\tTime: {tsdistances_times[i, j, 0]:.4f} (s), {tsdistances_times[i, j, 1]:.4f} (p), {tsdistances_times[i, j, 2]:.4f} (gpu) | AEON: {aeon_times[i, j]:.4f}")
            if not np.allclose(D, D_par):
                print("Parallel and single-threaded results do not match")

            if not np.allclose(D, D_aeon):
                print("AEON and tsdistances results do not match")

            np.save("times_tsdistances.npy", tsdistances_times)
            np.save("times_aeon.npy", aeon_times)


from typing import List, Optional, Union
from typeguard import TypeCheckError, typechecked, check_type
from tsdistances import tsdistances as tsd
import numpy as np


def check_input(
    u: np.ndarray, v: Optional[np.ndarray] = None
) -> Union[np.ndarray, Optional[np.ndarray]]:
    if u.ndim == 1:
        if v is None:
            raise ValueError("If `u` is 1-D, `v` must be not None.")
        if v.ndim == 1:
            _u = u.reshape((1, u.shape[0]))
            _v = v.reshape((1, v.shape[0]))
            return _u, _v
        elif v.ndim == 2:
            _u = u.reshape((1, u.shape[0]))
            _v = v
            return _u, _v
        else:
            raise ValueError("`v` must be 1-D or 2-D.")
    elif u.ndim == 2:
        _u = u
        if v is None:
            return _u, None
        else:
            if v.ndim == 1:
                _v = v.reshape((1, v.shape[0]))
                return _u, _v
            elif v.ndim == 2:
                _v = v
                return _u, _v
            else:
                raise ValueError("`v` must be 1-D or 2-D.")
    else:
        raise ValueError("`u` must be 1-D or 2-D.")


@typechecked
def euclidean_distance(
    u: np.ndarray, v: Optional[np.ndarray] = None, par: Optional[int] = 1
) -> Union[np.ndarray, float]:
    """
    Computes the Euclidean distance between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise Euclidean distances within `u`.

    Parameters
    ----------
    u : (N,) array_like or (M, N) array_like
        Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N) array_like, optional
    Input array. If provided, `v` should have the same shape as `u`.
    If `v` is None, pairwise distances within `u` are computed.
    par : int, optional
        Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
        The Euclidean distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> euclidean_distance([1, 0, 0], [0, 1, 0])
    1.4142135623730951
    >>> euclidean_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[1.41421356, 2.44948974],
           [1.        , 1.73205081]])
    >>> euclidean_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.        , 1.        ],
           [1.        , 0.        ]])

    """
    if u.ndim == 1 and v.ndim == 1:
        _u = u.reshape((1, u.shape[0]))
        _v = v.reshape((1, v.shape[0]))

    if u.ndim == 2:
        _u = u
        if v is None:
            return np.array(tsd.euclidean(_u, None, par))
        if v.ndim == 2:
            _v = v

    return np.array(tsd.euclidean(_u, _v, par))


@typechecked
def catcheucl_distance(
    u: np.ndarray, v: Optional[np.ndarray] = None, par: Optional[int] = 1
) -> Union[np.ndarray, float]:
    """
    Computes the Catch22-Euclidean distance between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise Catch22-Euclidean distances within `u`.

    Parameters
    ----------
    u : (N,) array_like or (M, N) array_like
        Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N) array_like, optional
    Input array. If provided, `v` should have the same shape as `u`.
    If `v` is None, pairwise distances within `u` are computed.
    par : int, optional
        Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
        The Catch22-Euclidean distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> catcheucl_distance([1, 0, 0], [0, 1, 0])
    1.0
    >>> catcheucl_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[1.0, 2.0], [1.0, 1.0]])
    >>> catcheucl_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 1.0], [1.0, 0.0]])

    """
    if u.ndim == 1 and v.ndim == 1:
        _u = u.reshape((1, u.shape[0]))
        _v = v.reshape((1, v.shape[0]))

    if u.ndim == 2:
        _u = u
        if v is None:
            return np.array(tsd.catch_euclidean(_u, None, par))
        if v.ndim == 2:
            _v = v

    return np.array(tsd.catch_euclidean(_u, _v, par))


@typechecked
def erp_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    gap_penalty: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Edit Distance with Real Penalty (ERP) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise ERP distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Chen, L. et al., On The Marriage of Lp-norms and Edit Distance, 2004.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    gap_penalty : double, optional
    Penalty for gap insertion/deletion (default is 0.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
        The ERP distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> erp_distance([1, 0, 0], [0, 1, 0])
    2.0
    >>> erp_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[2.0, 4.0], [1.0, 3.0]])
    >>> erp_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 1.0], [1.0, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.erp(_u, _v, band, gap_penalty, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.erp(_u, _v, band, gap_penalty, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.erp(_u, _v, band, gap_penalty, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.erp(_u, _v, band, gap_penalty, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.erp(_u, _v, band, gap_penalty, par, device))


@typechecked
def lcss_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    epsilon: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Longest Common Subsequence (LCSS) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise LCSS distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Vlachos, M. et al., Discovering Similar Multidimensional Trajectories, 2002.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    epsilon : double, optional
    Threshold value for the distance between two elements (default is 1.0).
    par : int, optional

    Returns
    -------
    distance : double or ndarray
        The LCSS distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> lcss_distance([1, 0, 0], [0, 1, 0], epsilon=0.5)
    0.3333333333333333
    >>> lcss_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]], epsilon=0.5)
    array([[0.3333333333333333, 0.6666666666666666], [0.0, 0.3333333333333333]])
    >>> lcss_distance([[1, 1, 1], [0, 1, 1]], epsilon=0.5)
    array([[0.0, 0.3333333333333333], [0.3333333333333333, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.lcss(_u, _v, band, epsilon, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.lcss(_u, _v, band, epsilon, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.lcss(_u, _v, band, epsilon, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.lcss(_u, _v, band, epsilon, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.lcss(_u, _v, band, epsilon, par, device))


@typechecked
def dtw_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Dynamic Time Warping (DTW) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise DTW distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Berndt, D.J. and Clifford, J., Using Dynamic Time Warping to Find Patterns in Time Series, 1994.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The DTW distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> dtw_distance([1, 0, 0], [0, 1, 0])
    1.0
    >>> dtw_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[2.0, 6.0], [1.0, 3.0]])
    >>> dtw_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.        , 1.        ], [1.        , 0.        ]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.dtw(_u, _v, band, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.dtw(_u, _v, band, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.dtw(_u, _v, band, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.dtw(_u, _v, band, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.dtw(_u, _v, band, par, device))


@typechecked
def ddtw_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Derivative Dynamic Time Warping (DDTW) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise DDTW distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Keogh, E. et al., Derivative Dynamic Time Warping, 2001.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The DDTW distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> ddtw_distance([1, 0, 0], [0, 1, 0])
    4.6875
    >>> ddtw_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[0.75, 1.6875], [0.1875, 0.0]])
    >>> ddtw_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 1.6875], [1.6875, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.ddtw(_u, _v, band, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.ddtw(_u, _v, band, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.ddtw(_u, _v, band, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.ddtw(_u, _v, band, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.ddtw(_u, _v, band, par, device))


@typechecked
def wdtw_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    g: Optional[float] = 0.05,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Weighted Dynamic Time Warping (WDTW) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise WDTW distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Jeong Y.-S. et al., Weighted dynamic time warping for time series classification, 2011.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The WDTW distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> wdtw_distance([1, 0, 0], [0, 1, 0])
    0.18242552380635635
    >>> wdtw_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[0.3648510476127127, 1.094553142838138], [0.18242552380635635, 0.547276571419069]])
    >>> wdtw_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 0.18242552380635635], [0.18242552380635635, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.wdtw(_u, _v, band, g, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.wdtw(_u, _v, band, g, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.wdtw(_u, _v, band, g, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.wdtw(_u, _v, band, g, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.wdtw(_u, _v, band, g, par, device))


@typechecked
def wddtw_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    g: Optional[float] = 0.05,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Weighted Derivative Dynamic Time Warping (WDDTW) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise WDDTW distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Jeong, Y.-S. et al., Weighted dynamic time warping for time series classification, 2011.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    par : int, optional

    Returns
    -------
    distance : double or ndarray
    The WDDTW distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> wddtw_distance([1, 0, 0], [0, 1, 0])
    0.8551196428422955
    >>> wddtw_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[0.13681914285476726, 0.3078430714232263], [0.034204785713691815, 0.0]])
    >>> wddtw_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 0.3078430714232263], [0.3078430714232263, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.wddtw(_u, _v, band, g, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.wddtw(_u, _v, band, g, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.wddtw(_u, _v, band, g, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.wddtw(_u, _v, band, g, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.wddtw(_u, _v, band, g, par, device))


@typechecked
def adtw_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    warp_penalty: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Amercing Dynamic Time Warping (ADTW) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise ADTW distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Hermann, M. et al., Amercing: An intuitive and effective constraint for dynamic time warping, 2023

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    w : double, optional
    Weight amercing penalty (default is 0.1).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The ADTW distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> adtw_distance([1, 0, 0], [0, 1, 0])
    0.0
    >>> adtw_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[0.0, 0.0], [0.0, 0.0]])
    >>> adtw_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 0.0], [0.0, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.adtw(_u, _v, band, warp_penalty, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.adtw(_u, _v, band, warp_penalty, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.adtw(_u, _v, band, warp_penalty, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.adtw(_u, _v, band, warp_penalty, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.adtw(_u, _v, band, warp_penalty, par, device))


@typechecked
def msm_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Move-Split-Merge (MSM) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise MSM distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Stefan, A. et al., The Move-Split-Merge Metric for Time Series, 2012.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The MSM distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> msm_distance([1, 0, 0], [0, 1, 0])
    2.0
    >>> msm_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[2.0, 4.0], [1.0, 3.0]])
    >>> msm_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 1.0], [1.0, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.msm(_u, _v, band, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.msm(_u, _v, band, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.msm(_u, _v, band, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.msm(_u, _v, band, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.msm(_u, _v, band, par, device))


@typechecked
def twe_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    band: Optional[float] = 1.0,
    stifness: Optional[float] = 0.001,
    penalty: Optional[float] = 1.0,
    par: Optional[bool] = True,
    device: Optional[str] = "cpu",
) -> Union[np.ndarray, float]:
    """
    Computes the Time Warp Edit (TWE) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise TWE distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Marteau, P., Time Warp Edit Distance with Stiffness Adjustment for Time Series Matching, 2008.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    band : double, optional
    Band size for the Sakoe-Chiba dynamic programming algorithm (default is 1.0).
    stifness : double, optional
    Elasticity parameter (default is 0.001).
    penalty : double, optional
    Penalty for gap insertion/deletion (default is 1.0).
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The TWE distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> twe_distance([1, 0, 0], [0, 1, 0])
    4.0
    >>> twe_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[3.0, 7.0], [1.0, 5.0]])
    >>> twe_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.0, 2.0], [2.0, 0.0]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.twe(_u, _v, band, stifness, penalty, par, device)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.twe(_u, _v, band, stifness, penalty, par, device))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.twe(_u, _v, band, stifness, penalty, par, device))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.twe(_u, _v, band, stifness, penalty, par, device))
            elif _v.shape[0] >= 2:
                return np.array(tsd.twe(_u, _v, band, stifness, penalty, par, device))


@typechecked
def sb_distance(
    u: np.ndarray,
    v: Optional[np.ndarray] = None,
    par: Optional[bool] = True,
) -> Union[np.ndarray, float]:
    """
    Computes the Shape-Based Distance (SBD) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise SBD distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Paparrizos, J. et al., k-Shape: Efficient and Accurate Clustering of Time Series, 2015.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.
    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.
    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The SBD distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> sb_distance([1, 0, 0], [0, 1, 0])
    1.4142135623730951
    >>> sb_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[1.41421356, 2.44948974], [1.        , 1.73205081]])
    >>> sb_distance([[1, 1, 1], [0, 1, 1]])
    array([[0.        , 1.        ], [1.        , 0.        ]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.sb(_u, _v, par)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.sb(_u, _v, par))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.sb(_u, _v, par))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.sb(_u, _v, par))
            elif _v.shape[0] >= 2:
                return np.array(tsd.sb(_u, _v, par))


def mp_distance(
    u: np.ndarray,
    window: Optional[int] = 20,
    v: Optional[np.ndarray] = None,
    par: Optional[bool] = True,
):
    """
    Computes the Matrix Profile distance (MPdist) [1] between two 1-D arrays or between two sets of 1-D arrays.
    If `v` is None, the function computes the pairwise MP distances within `u`.
    The length of the input arrays are not required to be the same.

    [1] Gharghabi S. et al., An Ultra-Fast Time Series Distance Measure to allow Data Mining in more Complex Real-World Deployments, 2020.

    Parameters
    ----------
    u : (N,) array_like or (M, N)
    Input array. If 1-D, `u` represents a single vector. If 2-D, `u` represents a set of vectors.

    v : (N,) array_like or (M, N), optional
    Input array.
    If `v` is None, pairwise distances within `u` are computed.

    window : int, optional
    Window size for the Matrix Profile calculation (default is 1).

    par : int, optional
    Number of jobs to use for computation (default is 1).

    Returns
    -------
    distance : double or ndarray
    The MP distance(s) between vectors/sets `u` and `v`.

    Examples
    --------
    >>> mp_distance([1, 0, 0], [0, 1, 0])
    1.4142135623730951
    >>> mp_distance([[1, 1, 1], [0, 1, 1]], [[0, 1, 0], [-1, 0, 0]])
    array([[1.41421356, 2.44948974], [1.        , 1.73205081]])
    >>> mp_distance([[1, 1, 1], [0, 1, 1]])
    p array([[0.        , 1.        ], [1.        , 0.        ]])
    """
    _u, _v = check_input(u, v)

    if _u.shape[0] == 1:
        if _v.shape[0] == 1:
            return tsd.mp(_u, window, _v, par)[0][0]
        elif v.shape[0] >= 2:
            return np.array(tsd.mp(_u, window, _v, par))
    elif _u.shape[0] >= 2:
        if _v is None:
            return np.array(tsd.mp(_u, window, _v, par))
        else:
            if _v.shape[0] == 1:
                return np.array(tsd.mp(_u, window, _v, par))
            elif _v.shape[0] >= 2:
                return np.array(tsd.mp(_u, window, _v, par))


__all__ = [
    "euclidean",
    "catch_euclidean",
    "erp",
    "lcss",
    "dtw",
    "ddtw",
    "wdtw",
    "wddtw",
    "adtw",
    "msm",
    "twe",
    "sb",
    "mp",
]



import pytest
import numpy as np
from tsdistances import (
    euclidean_distance,
    erp_distance,
    lcss_distance,
    dtw_distance,
    ddtw_distance,
    wdtw_distance,
    wddtw_distance,
    adtw_distance,
    msm_distance,
    twe_distance,
    sb_distance,
    mp_distance,
)
from aeon import distances as aeon
import stumpy
import time

N_SAMPLES = 10
A = np.loadtxt("tests/ACSF1/ACSF1_TRAIN.tsv", delimiter="\t")[:N_SAMPLES, 1:]
B = np.loadtxt("tests/ACSF1/ACSF1_TEST.tsv", delimiter="\t")[-N_SAMPLES:, 1:]
band = 1.0


def test_euclidean_distance():
    D = euclidean_distance(A, B, par=True)
    aeon_D = aeon.euclidean_pairwise_distance(A, B)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_erp_distance():
    gap_penalty = 0.0
    D = erp_distance(A, B, gap_penalty=gap_penalty, band=band, par=True)
    aeon_D = aeon.erp_pairwise_distance(A, B, g=gap_penalty, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_lcss_distance():
    epsilon = 0.1
    D = lcss_distance(A, B, epsilon=epsilon, band=band, par=True)
    aeon_D = aeon.lcss_pairwise_distance(A, B, epsilon=epsilon, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_dtw_distance():
    D = dtw_distance(A, B, band=band, par=True)
    aeon_D = aeon.dtw_pairwise_distance(A, B, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_ddtw_distance():
    D = ddtw_distance(A, B, band=band, par=True)
    aeon_D = aeon.ddtw_pairwise_distance(A, B, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_wdtw_distance():
    g = 0.05
    D = wdtw_distance(A, B, g=g, band=band, par=True)
    aeon_D = aeon.wdtw_pairwise_distance(A, B, g=g, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_wddtw_distance():
    g = 0.05
    D = wddtw_distance(A, B, g=g, band=band, par=True)
    aeon_D = aeon.wddtw_pairwise_distance(A, B, g=g, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_adtw_distance():
    warp_penalty = 1.0
    D = adtw_distance(A, B, band=band, warp_penalty=warp_penalty, par=True)
    aeon_D = aeon.adtw_pairwise_distance(A, B, window=band, warp_penalty=warp_penalty)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_msm_distance():
    D = msm_distance(A, B, band=band, par=True)
    aeon_D = aeon.msm_pairwise_distance(A, B, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_twe_distance():
    stiffness = 0.1
    penalty = 0.1
    D = twe_distance(A, B, band=band, stifness=stiffness, penalty=penalty, par=True)
    aeon_D = aeon.twe_pairwise_distance(A, B, nu=stiffness, lmbda=penalty, window=band)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_sb_distance():
    D = sb_distance(A, B, par=True)
    aeon_D = aeon.sbd_pairwise_distance(A, B)
    assert np.allclose(D, aeon_D, atol=1e-8)


def test_mp_distance():
    window = int(0.1 * A.shape[1])
    D = mp_distance(A, window, B, par=True)
    D_stumpy = np.zeros_like(D)
    for i in range(A.shape[0]):
        for j in range(B.shape[0]):
            D_stumpy[i, j] = stumpy.mpdist(A[i], B[j], m=window)
    assert np.allclose(D, D_stumpy, atol=1e-8)



import pytest
import numpy as np
from tsdistances import (
    erp_distance,
    lcss_distance,
    dtw_distance,
    ddtw_distance,
    wdtw_distance,
    wddtw_distance,
    adtw_distance,
    msm_distance,
    twe_distance,
)
import time

A = np.loadtxt('tests/ACSF1/ACSF1_TRAIN.tsv', delimiter='\t')[:10, 1:]
B = np.loadtxt('tests/ACSF1/ACSF1_TEST.tsv', delimiter='\t')[:10, 1:]
band = 1.0

def test_erp_distance():
    gap_penalty = 0.0
    D = erp_distance(A, B, gap_penalty=gap_penalty, band=band, par=True)
    D_gpu = erp_distance(A, B, gap_penalty=gap_penalty, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    # diff = np.where(abs(D - D_gpu) > (1e-8 + 0.1 * abs(D_gpu)))
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_lcss_distance():
    epsilon = 0.1
    D = lcss_distance(A, B, epsilon=epsilon, band=band, par=True)
    D_gpu = lcss_distance(A, B, epsilon=epsilon, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_dtw_distance():
    D = dtw_distance(A, B, band=band, par=True)
    D_gpu = dtw_distance(A, B, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_ddtw_distance():
    D = ddtw_distance(A, B, band=band, par=True)
    D_gpu =  ddtw_distance(A, B, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_wdtw_distance():
    g = 0.05
    D = wdtw_distance(A, B, g=g, band=band, par=True)
    D_gpu = wdtw_distance(A, B, g=g, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_wddtw_distance():
    g = 0.05
    D = wddtw_distance(A, B, g=g, band=band, par=True)
    D_gpu = wddtw_distance(A, B, g=g, band=band, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_adtw_distance():
    warp_penalty = 1.0
    D = adtw_distance(A, B, band=band, warp_penalty=warp_penalty, par=True)
    D_gpu = adtw_distance(A, B, band=band, warp_penalty=warp_penalty, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_msm_distance():
    D = msm_distance(A, B, band=band, par=True)
    D_gpu = msm_distance(A, B, band=band, par=True, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)


def test_twe_distance():
    stiffness = 0.1
    penalty = 0.1
    D = twe_distance(A, B, band=band, stifness=stiffness, penalty=penalty, par=True)
    D_gpu = twe_distance(A, B, band=band, stifness=stiffness, penalty=penalty, device='gpu')
    # Check that the GPU and CPU results are close (compare double precision with the single precision of GPU)
    assert np.allclose(D, D_gpu, rtol=0.1)